# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

from __future__ import absolute_import

import json
import os
import subprocess
from dataclasses import dataclass
from typing import Optional, Union
from functools import lru_cache

from clwpos import gettext as _
from clwpos.cl_wpos_exceptions import WposError, WpCliCommandError
from clwpos.constants import RedisRequiredConstants, WP_CLI
from clwpos.data_collector_utils import php_info
from clwpos.logsetup import (
    setup_logging,
    ADMIN_LOGFILE_PATH,
    USER_LOGFILE_PATH
)
from clwpos.utils import (
    WposUser,
    is_run_under_user,
    PHP,
    wp_cli_compatibility_check,
    run_in_cagefs_if_needed,
    user_name,
    litespeed_is_running
)

_logger = setup_logging(__name__)


@dataclass
class WordpressError:
    message: str
    context: dict


def get_php_version(abs_wp_path: str) -> PHP:
    """
    Return php_version that will be used for calling wp-cli commands.
    If 'CLWPOS_USE_SAVED_PHP_VERSION' envar is defined,
    try to get this version from a previously saved file.
    """
    use_saved_php_version = bool(os.environ.get("CLWPOS_USE_SAVED_PHP_VERSION"))
    if use_saved_php_version:
        php_version = _get_saved_php_version(abs_wp_path) or _get_php_version(abs_wp_path)
    else:
        php_version = _get_php_version(abs_wp_path)
    return php_version

@lru_cache(maxsize=None)
def _get_php_version(abs_wp_path: str) -> PHP:
    """Return PHP version."""
    result = php_info()

    items = []
    for item in result:
        if abs_wp_path.startswith(item["documentroot"]):
            items.append((item["documentroot"], item["version"]))
    items.sort(reverse=True)
    return items[0][1]


def _get_saved_php_version(abs_wp_path: str) -> Optional[PHP]:
    """
    Get domain's php version from a previously saved file.
    """
    if not is_run_under_user():
        raise WposError('Internal Error. Contact CloudLinux support')
    php_file_id = os.environ.get("CLWPOS_PHP_FILE_ID")
    php_info_file = WposUser(user_name()).php_info.format(file_id=php_file_id)
    if not os.path.exists(php_info_file):
        return None
    try:
        with open(php_info_file) as f:
            _php_info = json.load(f)
    except (OSError, json.decoder.JSONDecodeError) as e:
        _logger.exception("Error during reading of \".php_info\" file: %s", e)
        return None

    php_versions = []
    for vhost_info in _php_info:
        if abs_wp_path.startswith(vhost_info["documentroot"]):
            php_versions.append((vhost_info["documentroot"], vhost_info["version"]))

    if not php_versions:
        return None

    return PHP(sorted(php_versions, reverse=True)[0][1])


def wordpress(path: str, command: str, subcommand: str, *args, env=None) -> Union[str, WordpressError]:
    """
    Helper to execute wp commands, for example
        wp --path=<path> plugin install redis-cache
        wp --path=<path> plugin activate redis-cache
        wp --path=<path> redis enable
        wp --path=<path> plugin deactivate redis-cache
        wp --path=<path> plugin uninstall redis-cache
    @return: stderr if error was happened.
    """
    php_version = get_php_version(path)
    php_bin_path = str(php_version.bin())
    if not os.path.exists(php_bin_path):
        _logger.exception("Error during wp-cli command execution \"%s\": "
                          "invalid path to binary file \"%s\"",
                          command, php_bin_path)
        return WordpressError(
            message=_("Error during resolving path to php binary file:\n"
                      "got non-existent path \"%(path)s\"."),
            context={"path": php_bin_path}
        )
    # [attention] compatibility check may raise WpCliUnsupportedException exception
    wp_cli_compatibility_check(php_bin_path)
    command_part = [command, subcommand, *args]
    full_command = [
        WP_CLI,
        php_bin_path,
        path,
        *command_part
    ]
    environment = env or {}
    try:
        output = run_in_cagefs_if_needed(full_command, check=True, env=environment)
    except subprocess.CalledProcessError as error:
        post_check_flag = 'Post-check failed' in error.stderr
        command = ' '.join(full_command)
        _logger.exception("Error during command execution: \n%s\n"
                          "stdout=%s\n"
                          "stderr=%s",
                          command, error.stdout, error.stderr)
        logger_path = ADMIN_LOGFILE_PATH if not os.getuid() else USER_LOGFILE_PATH.format(homedir='HOME')
        if post_check_flag:
            message = _("WordPress plugin failed to activate correctly. Changes were reverted and caching module is now disabled.\n"
                      "Error reported from plugin: %(error_desc)s.\n"
                      "Event is logged to file: '%(logger_path)s' with stdout and stderr recorded.")
            context = {
                "command": command,
                "logger_path": logger_path,
                "error_desc": error.stdout
            }
        else:
            message = _("Unexpected error happened during command execution: '%(command)s'.\n"
                      "Event is logged to file: '%(logger_path)s' with stdout and stderr recorded.")
            context = {
                "command": command,
                "logger_path": logger_path
            }
        return WordpressError(
            message=message,
            context=context
        )

    return output.stdout


def is_multisite(path: str) -> bool:
    marker = 'cl_multisite_detected'
    command = 'if ( is_multisite() ) { echo "%s"; }' % marker
    result = wordpress(path, 'eval', command)
    if isinstance(result, WordpressError):
        raise WposError(message=result.message, context=result.context)
    return marker in result


def wp_get_constant(wp_path: str, constant: str, raise_exception=False) -> Optional[str]:
    """
    Get:
    - defined constant value
    - None in case of error
    - empty string if no such constant found
    """
    command = "if (defined('%(const)s')) { echo %(const)s; }" % {'const': constant}
    result = wordpress(wp_path, 'eval', command)
    if isinstance(result, WordpressError):
        if raise_exception:
            raise WpCliCommandError(message=result.message, context=result.context)
        _logger.error('Error during get WP constant: %s', result)
        return None
    return result


def diagnose_redis_connection_constants(docroot: str, wordpress_path: str):
    """
    Check required constants for redis connection establishment
    """
    redis_schema = wp_get_constant(os.path.join(docroot, wordpress_path), RedisRequiredConstants.WP_REDIS_SCHEME.name,
                                   raise_exception=True)
    if not redis_schema and redis_schema != RedisRequiredConstants.WP_REDIS_SCHEME.val:
        raise WposError('WordPress constant "%(constant)s" is not defined or defined with wrong value %(value)s',
                        context={'constant': RedisRequiredConstants.WP_REDIS_SCHEME.name, 'value': redis_schema})
    socket = wp_get_constant(os.path.join(docroot, wordpress_path), RedisRequiredConstants.WP_REDIS_PATH.name,
                             raise_exception=True)
    if not socket:
        raise WposError('WordPress constant "%(constant)s" is not defined',
                        context={'constant': RedisRequiredConstants.WP_REDIS_PATH.name})
    if not os.path.exists(socket):
        raise WposError('Redis socket %(socket)s does not exist in the system',
                        context={'socket': socket})


def obtain_wp_cli_env(abs_wp_path: str):
    """
    Returns needed envars for wp-cli
    """
    env = None
    if os.path.exists(os.path.join(abs_wp_path, '.htaccess')):
        server_software = 'LiteSpeed' if litespeed_is_running() else 'Apache'
        env = {'SERVER_SOFTWARE': server_software}
    return env


def is_plugin_activated(abs_wp_path: str, plugin_name: str, wp_cli_env=None) -> bool:
    result = wordpress(abs_wp_path, "plugin", "is-active", plugin_name, env=wp_cli_env)
    return not isinstance(result, WordpressError)


def is_plugin_installed(abs_wp_path: str, plugin_name: str, wp_cli_env=None) -> bool:
    result = wordpress(abs_wp_path, "plugin", "is-installed", plugin_name, env=wp_cli_env)
    return not isinstance(result, WordpressError)
